@@ -27,20 +27,19 @@
 #include <stdarg.h>
 #include "windef.h"
 #include "winbase.h"
+
+#include "wine/list.h"
 #include "wine/unixlib.h"
 
-#define SUCCESS(ret) (ret == SQL_SUCCESS || ret == SQL_SUCCESS_WITH_INFO)
+static inline BOOL SUCCESS( SQLRETURN ret ) { return ret == SQL_SUCCESS || ret == SQL_SUCCESS_WITH_INFO; }
 
 enum sql_funcs
 {
     process_attach,
-    unix_SQLAllocConnect,
-    unix_SQLAllocEnv,
+    process_detach,
     unix_SQLAllocHandle,
     unix_SQLAllocHandleStd,
-    unix_SQLAllocStmt,
     unix_SQLBindCol,
-    unix_SQLBindParam,
     unix_SQLBindParameter,
     unix_SQLBrowseConnect,
     unix_SQLBrowseConnectW,
@@ -58,16 +57,12 @@ enum sql_funcs
     unix_SQLConnect,
     unix_SQLConnectW,
     unix_SQLCopyDesc,
-    unix_SQLDataSources,
-    unix_SQLDataSourcesW,
     unix_SQLDescribeCol,
     unix_SQLDescribeColW,
     unix_SQLDescribeParam,
     unix_SQLDisconnect,
     unix_SQLDriverConnect,
     unix_SQLDriverConnectW,
-    unix_SQLDrivers,
-    unix_SQLDriversW,
     unix_SQLEndTran,
     unix_SQLError,
     unix_SQLErrorW,
@@ -79,8 +74,6 @@ enum sql_funcs
     unix_SQLFetchScroll,
     unix_SQLForeignKeys,
     unix_SQLForeignKeysW,
-    unix_SQLFreeConnect,
-    unix_SQLFreeEnv,
     unix_SQLFreeHandle,
     unix_SQLFreeStmt,
     unix_SQLGetConnectAttr,
@@ -152,10 +145,34 @@ enum sql_funcs
     unix_funcs_count
 };
 
+struct bind_col_args
+{
+    INT16 target_type;
+    void *target_value;
+    INT64 buffer_length;
+};
+
+struct bind_parameter_args
+{
+    INT16  input_output_type;
+    INT16  value_type;
+    INT16  parameter_type;
+    UINT64 column_size;
+    INT16  decimal_digits;
+    void  *parameter_value;
+    INT64  buffer_length;
+};
+
 struct param
 {
-    INT64 len;  /* result length stored in Unix lib */
-    void *ptr;  /* result length ptr passed by client */
+    INT16 type;
+    union
+    {
+        struct bind_col_args col;
+        struct bind_parameter_args parameter;
+    };
+    UINT8 *len;  /* result length array stored in Unix lib */
+    void  *ptr;  /* result length ptr passed by client */
 };
 
 struct param_binding
@@ -164,43 +181,69 @@ struct param_binding
     struct param *param;
 };
 
-struct handle
+struct object
 {
+    UINT32 type;
     UINT64 unix_handle;
+    void  *win32_handle;
+    const struct win32_funcs *win32_funcs;
+    struct object *parent;
+    struct list entry;
+    struct list children;
+    CRITICAL_SECTION cs;
+    BOOL closed;
+};
+
+struct environment
+{
+    struct object hdr;
+    /* attributes */
+    UINT32 attr_version;
+    /* drivers and data sources */
+    UINT32 drivers_idx;
+    void  *drivers_key;
+    UINT32 sources_idx;
+    void  *sources_key;
+    BOOL   sources_system;
+};
+
+struct connection
+{
+    struct object hdr;
+    UINT32 driver_odbc_ver;
+    /* attributes */
+    UINT32 attr_con_timeout;
+    UINT32 attr_login_timeout;
+};
+
+struct statement
+{
+    struct object hdr;
+    /* descriptors */
+    struct descriptor *desc[4];
+    /* parameter bindings */
     struct param_binding bind_col;
-    struct param_binding bind_param;
     struct param_binding bind_parameter;
+    UINT32 row_count;   /* number of rows returned by SQLFetch() */
 };
 
-struct SQLAllocConnect_params
+struct descriptor
 {
-    UINT64 EnvironmentHandle;
-    UINT64 ConnectionHandle;
-};
-
-struct SQLAllocEnv_params
-{
-    UINT64 EnvironmentHandle;
+    struct object hdr;
 };
 
 struct SQLAllocHandle_params
 {
-    INT16  HandleType;
-    UINT64 InputHandle;
-    UINT64 OutputHandle;
+    INT16   HandleType;
+    UINT64  InputHandle;
+    UINT64 *OutputHandle;
 };
 
 struct SQLAllocHandleStd_params
 {
-    INT16  HandleType;
-    UINT64 InputHandle;
-    UINT64 OutputHandle;
-};
-
-struct SQLAllocStmt_params
-{
-    UINT64 ConnectionHandle;
-    UINT64 StatementHandle;
+    INT16   HandleType;
+    UINT64  InputHandle;
+    UINT64 *OutputHandle;
 };
 
 struct SQLBindCol_params
@@ -210,19 +253,7 @@ struct SQLBindCol_params
     INT16  TargetType;
     void  *TargetValue;
     INT64  BufferLength;
-    INT64 *StrLen_or_Ind;
-};
-
-struct SQLBindParam_params
-{
-    UINT64 StatementHandle;
-    UINT16 ParameterNumber;
-    INT16  ValueType;
-    INT16  ParameterType;
-    UINT64 LengthPrecision;
-    INT16  ParameterScale;
-    void  *ParameterValue;
-    INT64 *StrLen_or_Ind;
+    void  *StrLen_or_Ind;
 };
 
 struct SQLBindParameter_params
@@ -236,7 +267,7 @@ struct SQLBindParameter_params
     INT16  DecimalDigits;
     void  *ParameterValue;
     INT64  BufferLength;
-    INT64 *StrLen_or_Ind;
+    void  *StrLen_or_Ind;
 };
 
 struct SQLBrowseConnect_params
@@ -399,30 +430,6 @@ struct SQLCopyDesc_params
     UINT64 TargetDescHandle;
 };
 
-struct SQLDataSources_params
-{
-    UINT64 EnvironmentHandle;
-    UINT16 Direction;
-    UCHAR *ServerName;
-    INT16  BufferLength1;
-    INT16 *NameLength1;
-    UCHAR *Description;
-    INT16  BufferLength2;
-    INT16 *NameLength2;
-};
-
-struct SQLDataSourcesW_params
-{
-    UINT64 EnvironmentHandle;
-    UINT16 Direction;
-    WCHAR *ServerName;
-    INT16  BufferLength1;
-    INT16 *NameLength1;
-    WCHAR *Description;
-    INT16  BufferLength2;
-    INT16 *NameLength2;
-};
-
 struct SQLDescribeCol_params
 {
     UINT64  StatementHandle;
@@ -488,30 +495,6 @@ struct SQLDriverConnectW_params
     UINT16 DriverCompletion;
 };
 
-struct SQLDrivers_params
-{
-    UINT64 EnvironmentHandle;
-    UINT16 Direction;
-    UCHAR *DriverDescription;
-    INT16  BufferLength1;
-    INT16 *DescriptionLength;
-    UCHAR *DriverAttributes;
-    INT16  BufferLength2;
-    INT16 *AttributesLength;
-};
-
-struct SQLDriversW_params
-{
-    UINT64 EnvironmentHandle;
-    UINT16 Direction;
-    WCHAR *DriverDescription;
-    INT16  BufferLength1;
-    INT16 *DescriptionLength;
-    WCHAR *DriverAttributes;
-    INT16  BufferLength2;
-    INT16 *AttributesLength;
-};
-
 struct SQLEndTran_params
 {
     INT16  HandleType;
@@ -617,16 +600,6 @@ struct SQLForeignKeysW_params
     INT16  NameLength6;
 };
 
-struct SQLFreeConnect_params
-{
-    UINT64 ConnectionHandle;
-};
-
-struct SQLFreeEnv_params
-{
-    UINT64 EnvironmentHandle;
-};
-
 struct SQLFreeHandle_params
 {
     INT16  HandleType;
